@@ -34,6 +34,7 @@ def _on_error(response):
 
   """
   if response.status_code == 401:
+    _logger.error(response.content)
     raise HdfsError('Authentication failure. Check your credentials.')
   try:
     # Cf. http://hadoop.apache.org/docs/r1.0.4/webhdfs.html#Error+Responses
@@ -78,9 +79,11 @@ class _Request(object):
       url = '%s%s%s' % (
         client.url.rstrip('/'),
         self.webhdfs_prefix,
-        client.resolve(hdfs_path),
+        quote(client.resolve(hdfs_path), '/= '),
       )
       params['op'] = operation
+      if client._proxy is not None:
+        params['doas'] = client._proxy
       return client._request(
         method=self.method,
         url=url,
@@ -149,10 +152,7 @@ class Client(object):
     self.root = root
     self.url = url
     self._session = session or rq.Session()
-    if proxy:
-      if not self._session.params:
-        self._session.params = {}
-      self._session.params['doas'] = proxy
+    self._proxy = proxy
     self._timeout = timeout
     _logger.info('Instantiated %r.', self)
 
@@ -185,19 +185,20 @@ class Client(object):
 
   # Raw API endpoints
 
-  _append = _Request('PUT') # doesn't allow for streaming
-  _append_1 = _Request('POST', allow_redirects=False) # cf. `read`
-  _create = _Request('PUT') # doesn't allow for streaming
-  _create_1 = _Request('PUT', allow_redirects=False) # cf. `write`
+  _append = _Request('POST', allow_redirects=False) # cf. `read`
+  _create = _Request('PUT', allow_redirects=False) # cf. `write`
   _delete = _Request('DELETE')
+  _get_acl_status = _Request('GET')
   _get_content_summary = _Request('GET')
   _get_file_checksum = _Request('GET')
   _get_file_status = _Request('GET')
   _get_home_directory = _Request('GET')
   _list_status = _Request('GET')
   _mkdirs = _Request('PUT')
+  _modify_acl_entries = _Request('PUT')
   _open = _Request('GET', stream=True)
   _rename = _Request('PUT')
+  _set_acl = _Request('PUT')
   _set_owner = _Request('PUT')
   _set_permission = _Request('PUT')
   _set_replication = _Request('PUT')
@@ -248,7 +249,7 @@ class Client(object):
     # #LATEST expansion (could cache the pattern, but not worth it)
 
     _logger.debug('Resolved path %r to %r.', hdfs_path, path)
-    return quote(path, '/= ')
+    return path
 
   def content(self, hdfs_path, strict=True):
     """Get ContentSummary_ for a file or folder on HDFS.
@@ -280,6 +281,47 @@ class Client(object):
     res = self._get_file_status(hdfs_path, strict=strict)
     return res.json()['FileStatus'] if res else None
 
+  def acl_status(self, hdfs_path, strict=True):
+    """Get AclStatus_ for a file or folder on HDFS.
+
+    :param hdfs_path: Remote path.
+    :param strict: If `False`, return `None` rather than raise an exception if
+      the path doesn't exist.
+
+    .. _AclStatus: ACLS_
+    .. _ACLS: https://hadoop.apache.org/docs/stable2/hadoop-project-dist/hadoop-hdfs/WebHDFS.html#Get_ACL_Status
+
+    """
+    _logger.info('Fetching ACL status for %r.', hdfs_path)
+    res = self._get_acl_status(hdfs_path, strict=strict)
+    return res.json()['AclStatus'] if res else None
+
+  def set_acl(self, hdfs_path, acl_spec, clear=True):
+    """SetAcl_ or ModifyAcl_ for a file or folder on HDFS.
+
+    :param hdfs_path: Path to an existing remote file or directory. An
+      :class:`HdfsError` will be raised if the path doesn't exist.
+    :param acl_spec: String representation of an ACL spec. Must be a valid
+      string with entries for user, group and other. For example:
+      `"user::rwx,user:foo:rw-,group::r--,other::---"`.
+    :param clear: Clear existing ACL entries. If set to false, all existing ACL
+      entries that are not specified in this call are retained without changes,
+      behaving like ModifyAcl_. For example: `"user:foo:rwx"`.
+
+    .. _SetAcl: SETACL_
+    .. SETACL_: https://hadoop.apache.org/docs/stable2/hadoop-project-dist/hadoop-hdfs/WebHDFS.html#Set_ACL
+
+    .. ModifyAcl_: MODACL_
+    .. MODACL_: https://hadoop.apache.org/docs/stable2/hadoop-project-dist/hadoop-hdfs/WebHDFS.html#Modify_ACL_Entries
+
+    """
+    if clear:
+      _logger.info('Setting ACL spec for %r to %r.', hdfs_path, acl_spec)
+      self._set_acl(hdfs_path, aclspec=acl_spec)
+    else:
+      _logger.info('Modifying ACL spec for %r to %r.', hdfs_path, acl_spec)
+      self._modify_acl_entries(hdfs_path, aclspec=acl_spec)
+
   def parts(self, hdfs_path, parts=None, status=False):
     """Returns a dictionary of part-files corresponding to a path.
 
@@ -373,10 +415,10 @@ class Client(object):
       if permission or blocksize or replication:
         raise ValueError('Cannot change file properties while appending.')
       _logger.info('Appending to %r.', hdfs_path)
-      res = self._append_1(hdfs_path, buffersize=buffersize)
+      res = self._append(hdfs_path, buffersize=buffersize)
     else:
       _logger.info('Writing to %r.', hdfs_path)
-      res = self._create_1(
+      res = self._create(
         hdfs_path,
         overwrite=overwrite,
         permission=permission,
@@ -391,7 +433,6 @@ class Client(object):
         method='POST' if append else 'PUT',
         url=res.headers['location'],
         data=(c.encode(encoding) for c in _data) if encoding else _data,
-        auth=False,
       )
 
     if data is None:
@@ -400,7 +441,7 @@ class Client(object):
       consumer(data)
 
   def upload(self, hdfs_path, local_path, overwrite=False, n_threads=1,
-    temp_dir=None, chunk_size=2 ** 16, progress=None, **kwargs):
+    temp_dir=None, chunk_size=2 ** 16, progress=None, cleanup=True, **kwargs):
     """Upload a file or directory to HDFS.
 
     :param hdfs_path: Target HDFS path. If it already exists and is a
@@ -419,6 +460,8 @@ class Client(object):
       `chunk_size` bytes. It will be passed two arguments, the path to the
       file being uploaded and the number of bytes transferred so far. On
       completion, it will be called once with `-1` as second argument.
+    :param cleanup: Delete any uploaded files if an error occurs during the
+      upload.
     :param \*\*kwargs: Keyword arguments forwarded to :meth:`write`.
 
     On success, this method returns the remote upload path.
@@ -457,13 +500,17 @@ class Client(object):
     try:
       statuses = [status for _, status in self.list(hdfs_path, status=True)]
     except HdfsError as err:
-      if 'not a directory' in str(err):
+      message = str(err)
+      if 'not a directory' in message:
         # Remote path is a normal file.
         if not overwrite:
           raise HdfsError('Remote path %r already exists.', hdfs_path)
-      else:
+      elif 'does not exist' in message:
         # Remote path doesn't exist.
         temp_path = hdfs_path
+      else:
+        # An unexpected error occurred.
+        raise err
     else:
       # Remote path is a directory.
       suffixes = set(status['pathSuffix'] for status in statuses)
@@ -519,12 +566,15 @@ class Client(object):
       else:
         _map_async(n_threads, _upload, fpath_tuples)
     except Exception as err: # pylint: disable=broad-except
-      _logger.exception('Error while uploading. Attempting cleanup.')
-      try:
-        self.delete(temp_path, recursive=True)
-      except Exception:
-        _logger.error('Unable to cleanup temporary folder.')
-      finally:
+      if cleanup:
+        _logger.exception('Error while uploading. Attempting cleanup.')
+        try:
+          self.delete(temp_path, recursive=True)
+        except Exception:
+          _logger.error('Unable to cleanup temporary folder.')
+        finally:
+          raise err
+      else:
         raise err
     else:
       if temp_path != hdfs_path:
@@ -575,6 +625,14 @@ class Client(object):
 
     This ensures that connections are always properly closed.
 
+    .. note::
+
+      The raw file-like object returned by this method (when called without an
+      encoding, chunk size, or delimiter) can have a very different performance
+      profile than local files. In particular, line-oriented methods are often
+      slower. The recommended workaround is to specify an encoding when
+      possible or read the entire file before splitting it.
+
     """
     if progress and not chunk_size:
       raise ValueError('Progress callback requires a positive chunk size.')
@@ -683,7 +741,7 @@ class Client(object):
         raise HdfsError('Parent directory of %r does not exist.', local_path)
       temp_path = local_path
     # Then we figure out which files we need to download and where.
-    remote_paths = list(self.walk(hdfs_path))
+    remote_paths = list(self.walk(hdfs_path, depth=0, status=False))
     if not remote_paths:
       # This is a single file.
       remote_fpaths = [hdfs_path]
@@ -768,7 +826,8 @@ class Client(object):
     :param hdfs_src_path: Source path.
     :param hdfs_dst_path: Destination path. If the path already exists and is
       a directory, the source will be moved into it. If the path exists and is
-      a file, this method will raise an :class:`HdfsError`.
+      a file, or if a parent destination directory is missing, this method will
+      raise an :class:`HdfsError`.
 
     """
     _logger.info('Renaming %r to %r.', hdfs_src_path, hdfs_dst_path)
@@ -975,6 +1034,9 @@ class InsecureClient(Client):
     `whoami`).
   :param \*\*kwargs: Keyword arguments passed to the base class' constructor.
 
+  Note that if a session argument is passed in, it will be modified in-place to
+  support authentication.
+
   """
 
   def __init__(self, url, user=None, **kwargs):
@@ -995,6 +1057,9 @@ class TokenClient(Client):
   :param token: Hadoop delegation token.
   :param \*\*kwargs: Keyword arguments passed to the base class' constructor.
 
+  Note that if a session argument is passed in, it will be modified in-place to
+  support authentication.
+
   """
 
   def __init__(self, url, token, **kwargs):
